"""
Views to send private files.
"""
import os
from urllib.parse import quote

from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.utils.module_loading import import_string
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin

from . import appconfig
from .models import PrivateFile
from .servers import get_server_class
from .storage import private_storage


class PrivateStorageView(View):
    """
    Return the uploaded files
    """

    #: The storage class to retrieve files from
    storage = private_storage

    #: The authorisation rule for accessing
    can_access_file = staticmethod(import_string(appconfig.PRIVATE_STORAGE_AUTH_FUNCTION))

    #: Import the server class once
    server_class = get_server_class(appconfig.PRIVATE_STORAGE_SERVER)

    #: Whether the file should be displayed ``inline`` or show a download box (``attachment``).
    content_disposition = None

    #: The filename to use when :attr:`content_disposition` is set.
    content_disposition_filename = None

    #: Message to be displayed when the user cannot access the requested file.
    permission_denied_message = "Private storage access denied"

    def get_path(self):
        """
        Determine the path for the object to provide.
        This can be overwritten to combine the view with a different object retrieval.
        """
        return self.kwargs['path']

    def get_storage(self):
        """
        Tell which storage to retrieve the file from.
        """
        return self.storage

    def get_private_file(self):
        """
        Return all relevant data in a single object, so this is easy to extend
        and server implementations can pick what they need.
        """
        return PrivateFile(
            request=self.request,
            storage=self.get_storage(),
            relative_name=self.get_path()
        )

    def get(self, request, *args, **kwargs):
        """
        Handle incoming GET requests
        """
        private_file = self.get_private_file()

        if not self.can_access_file(private_file):
            raise PermissionDenied(self.permission_denied_message)

        if not private_file.exists():
            return self.serve_file_not_found(private_file)
        else:
            return self.serve_file(private_file)

    def serve_file_not_found(self, private_file):
        """
        Display a response message telling that the file is not found.
        This can be overwritten to improve the customer experience.
        For example
        - redirect the user, and show a message.
        - render the message in the expected media type (e.g. PNG).
        - show a custom 404 page.

        :type private_file: :class:`private_storage.models.PrivateFile`
        :rtype: django.http.HttpResponse
        """
        raise Http404("File not found")

    def serve_file(self, private_file):
        """
        Serve the file that was retrieved from the storage.
        The relative path can be found with ``private_file.relative_name``.

        :type private_file: :class:`private_storage.models.PrivateFile`
        :rtype: django.http.HttpResponse
        """
        response = self.server_class().serve(private_file)

        if self.content_disposition:
            # Join syntax works in all Python versions. Python 3 doesn't support b'..'.format(),
            # and % formatting was added for bytes in 3.5: https://bugs.python.org/issue3982
            filename = self.get_content_disposition_filename(private_file)
            response['Content-Disposition'] = b'; '.join([
                self.content_disposition.encode(), self._encode_filename_header(filename)
            ])

        return response

    def get_content_disposition_filename(self, private_file):
        """
        Return the filename in the download header.
        """
        return self.content_disposition_filename or os.path.basename(private_file.relative_name)

    def _encode_filename_header(self, filename):
        """
        The filename, encoded to use in a ``Content-Disposition`` header.
        """
        # Based on https://www.djangosnippets.org/snippets/1710/
        user_agent = self.request.headers.get('user-agent', '')
        if 'WebKit' in user_agent:
            # Support available for UTF-8 encoded strings.
            # This also matches Edgee.
            return f'filename={filename}'.encode("utf-8")
        elif 'MSIE' in user_agent:
            # IE does not support RFC2231 for internationalized headers, but somehow
            # percent-decodes it so this can be used instead. Note that using the word
            # "attachment" anywhere in the filename overrides an inline content-disposition.
            url_encoded = quote(filename.encode("utf-8")).replace('attachment', "a%74tachment")
            return f"filename={url_encoded}".encode("utf-8")
        else:
            # For others like Firefox, we follow RFC2231 (encoding extension in HTTP headers).
            rfc2231_filename = quote(filename.encode("utf-8"))
            return f"filename*=UTF-8''{rfc2231_filename}".encode("utf-8")


class PrivateStorageDetailView(SingleObjectMixin, PrivateStorageView):
    """
    Download a document based on an object ID.
    This view can by used by third-party apps to implement their own download view.

    Implement access controls by overriding :meth`get_queryset` or redefining :meth:`can_access_file`.
    """

    #: Define the model to fetch.
    model = None

    #: Define which field the file name is stored at.
    model_file_field = 'file'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def get_path(self):
        file = getattr(self.object, self.model_file_field)
        return file.name

    def get_storage(self):
        field = self.object._meta.get_field(self.model_file_field)
        return field.storage

    def get_private_file(self):
        # Provide the parent object as well.
        return PrivateFile(
            request=self.request,
            storage=self.get_storage(),
            relative_name=self.get_path(),
            parent_object=self.object
        )

    def can_access_file(self, private_file):
        """
        The authorization rule for this view.
        By default it reuses the ``PRIVATE_STORAGE_AUTH_FUNCTION`` setting,
        but this should likely be redefined.
        """
        return PrivateStorageView.can_access_file(private_file)



import datetime
import logging
import os
import posixpath
import warnings

from django.core import checks
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.db.models.fields.files import ImageFieldFile, ImageFileDescriptor
from django.forms import ImageField
from django.template.defaultfilters import filesizeformat
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from .storage import private_storage

logger = logging.getLogger(__name__)


class PrivateFileField(models.FileField):
    """
    Filefield with private storage, custom filename and size checks.

    Extra settings:
    - ``upload_subfolder``: a lambda to find the subfolder, depending on the instance.
    - ``content_types``: list of allowed content types.
    - ``max_file_size``: maximum file size.
    """
    default_error_messages = {
        'invalid_file_type': _('File type not supported.'),
        'file_too_large': _('The file may not be larger than {max_size}.'),
    }

    def __init__(self, *args, **kwargs):
        self.upload_subfolder = kwargs.pop('upload_subfolder', None)
        self.content_types = kwargs.pop("content_types", None) or ()
        self.max_file_size = kwargs.pop("max_file_size", None)

        kwargs.setdefault('storage', private_storage)
        super().__init__(*args, **kwargs)

    def clean(self, *args, **kwargs):
        data = super().clean(*args, **kwargs)
        file = data.file
        if isinstance(file, UploadedFile):
            # content_type is only available for uploaded files,
            # and not for files which are already stored in the model.
            content_type = file.content_type

            if self.content_types and content_type not in self.content_types:
                logger.debug('Rejected uploaded file type: %s', content_type)
                raise ValidationError(self.error_messages['invalid_file_type'])

            if self.max_file_size and file.size > self.max_file_size:
                raise ValidationError(self.error_messages['file_too_large'].format(
                    max_size=filesizeformat(self.max_file_size),
                    size=filesizeformat(file.size)
                ))

        return data

    def generate_filename(self, instance, filename):
        path_parts = []

        if self.upload_to:
            # Support the upload_to callable that Django provides
            if callable(self.upload_to):
                dirname, filename = os.path.split(self.upload_to(instance, filename))
                path_parts.append(dirname)
            else:
                dirname = force_str(datetime.datetime.now().strftime(force_str(self.upload_to)))
                path_parts.append(dirname)

        # Add our custom subdir function.
        upload_subfolder = self.upload_subfolder
        if upload_subfolder:
            # Should return a list, so joining can be done in a storage-specific manner.
            extra_dirs = upload_subfolder(instance)

            # Avoid mistakes by developers, no "s/u/b/p/a/t/h/"
            if isinstance(extra_dirs, str):
                warnings.warn("{}.{}.upload_subfolder should return a list"
                              " to avoid path-separator issues.".format(
                    instance.__class__.__name__, self.name), UserWarning)
                extra_dirs = os.path.split(extra_dirs)

            path_parts.extend([self.storage.get_valid_name(dir) for dir in extra_dirs])

        path_parts.append(self._get_clean_filename(filename))
        filename = posixpath.join(*path_parts)
        return self.storage.generate_filename(filename)

    def _get_clean_filename(self, filename):
        # As of Django 1.10+, file names are no longer cleaned locally, but cleaned by the storage.
        # This compatibility function makes sure all Django versions generate a safe filename.
        return os.path.normpath(self.storage.get_valid_name(os.path.basename(filename)))


class PrivateImageField(PrivateFileField):
    attr_class = ImageFieldFile
    descriptor_class = ImageFileDescriptor
    description = _("Image")

    def __init__(self, verbose_name=None, name=None, width_field=None, height_field=None, **kwargs):
        self.width_field, self.height_field = width_field, height_field
        super().__init__(verbose_name, name, **kwargs)

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_image_library_installed(),
        ]

    def _check_image_library_installed(self):
        try:
            from PIL import Image  # NOQA
        except ImportError:
            return [
                checks.Error(
                    'Cannot use ImageField because Pillow is not installed.',
                    hint=('Get Pillow at https://pypi.org/project/Pillow/ '
                          'or run command "python -m pip install Pillow".'),
                    obj=self,
                    id='fields.E210',
                )
            ]
        else:
            return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.width_field:
            kwargs['width_field'] = self.width_field
        if self.height_field:
            kwargs['height_field'] = self.height_field
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        # Attach update_dimension_fields so that dimension fields declared
        # after their corresponding image field don't stay cleared by
        # Model.__init__, see bug #11196.
        # Only run post-initialization dimension update on non-abstract models
        if not cls._meta.abstract:
            models.signals.post_init.connect(self.update_dimension_fields, sender=cls)

    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        """
        Update field's width and height fields, if defined.

        This method is hooked up to model's post_init signal to update
        dimensions after instantiating a model instance.  However, dimensions
        won't be updated if the dimensions fields are already populated.  This
        avoids unnecessary recalculation when loading an object from the
        database.

        Dimensions can be forced to update with force=True, which is how
        ImageFileDescriptor.__set__ calls this method.
        """
        # Nothing to update if the field doesn't have dimension fields or if
        # the field is deferred.
        has_dimension_fields = self.width_field or self.height_field
        if not has_dimension_fields or self.attname not in instance.__dict__:
            return

        # getattr will call the ImageFileDescriptor's __get__ method, which
        # coerces the assigned value into an instance of self.attr_class
        # (ImageFieldFile in this case).
        file = getattr(instance, self.attname)

        # Nothing to update if we have no file and not being forced to update.
        if not file and not force:
            return

        dimension_fields_filled = not(
            (self.width_field and not getattr(instance, self.width_field)) or
            (self.height_field and not getattr(instance, self.height_field))
        )
        # When both dimension fields have values, we are most likely loading
        # data from the database or updating an image field that already had
        # an image stored.  In the first case, we don't want to update the
        # dimension fields because we are already getting their values from the
        # database.  In the second case, we do want to update the dimensions
        # fields and will skip this return because force will be True since we
        # were called from ImageFileDescriptor.__set__.
        if dimension_fields_filled and not force:
            return

        # file should be an instance of ImageFieldFile or should be None.
        if file:
            width = file.width
            height = file.height
        else:
            # No file, so clear dimensions fields.
            width = None
            height = None

        # Update the width and height fields.
        if self.width_field:
            setattr(instance, self.width_field, width)
        if self.height_field:
            setattr(instance, self.height_field, height)

    def formfield(self, **kwargs):
        return super().formfield(**{
            'form_class': ImageField,
            **kwargs,
        })



try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from django.utils.deconstruct import deconstructible
from storages.backends.s3boto3 import S3Boto3Storage
from storages.utils import setting

from private_storage import appconfig


@deconstructible
class PrivateS3BotoStorage(S3Boto3Storage):
    """
    Private storage bucket for S3
    """

    #access_key_names = ['AWS_PRIVATE_S3_ACCESS_KEY_ID', 'AWS_PRIVATE_ACCESS_KEY_ID'] + getattr(S3Boto3Storage, 'access_key_names', [])
    #secret_key_names = ['AWS_PRIVATE_S3_SECRET_ACCESS_KEY', 'AWS_PRIVATE_SECRET_ACCESS_KEY'] + getattr(S3Boto3Storage, 'secret_key_names', [])

    def __init__(self, **settings):
        super().__init__(**settings)
        self.file_overwrite = setting('AWS_PRIVATE_S3_FILE_OVERWRITE', False)  # false, differ from base class
        self.object_parameters = setting('AWS_PRIVATE_S3_OBJECT_PARAMETERS', {})
        self.bucket_name = setting('AWS_PRIVATE_STORAGE_BUCKET_NAME')
        self.auto_create_bucket = setting('AWS_PRIVATE_AUTO_CREATE_BUCKET', False)
        self.default_acl = setting('AWS_PRIVATE_DEFAULT_ACL', 'private')  # differ from base class
        self.bucket_acl = setting('AWS_PRIVATE_BUCKET_ACL', self.default_acl)
        self.querystring_auth = setting('AWS_PRIVATE_QUERYSTRING_AUTH', True)
        self.querystring_expire = setting('AWS_PRIVATE_QUERYSTRING_EXPIRE', 3600)
        self.signature_version = setting('AWS_PRIVATE_S3_SIGNATURE_VERSION')
        self.reduced_redundancy = setting('AWS_PRIVATE_REDUCED_REDUNDANCY', False)
        self.location = setting('AWS_PRIVATE_LOCATION', '')
        self.encryption = setting('AWS_PRIVATE_S3_ENCRYPTION', False)
        self.custom_domain = setting('AWS_PRIVATE_S3_CUSTOM_DOMAIN')
        self.addressing_style = setting('AWS_PRIVATE_S3_ADDRESSING_STYLE')
        self.secure_urls = setting('AWS_PRIVATE_S3_SECURE_URLS', True)
        self.file_name_charset = setting('AWS_PRIVATE_S3_FILE_NAME_CHARSET', 'utf-8')
        self.preload_metadata = setting('AWS_PRIVATE_PRELOAD_METADATA', False)
        self.endpoint_url = setting('AWS_PRIVATE_S3_ENDPOINT_URL', None)
        self.use_ssl = setting('AWS_PRIVATE_S3_USE_SSL', True)

        # default settings used to be class attributes on S3Boto3Storage, but
        # are now part of the initialization or moved to a dictionary
        self.access_key = setting('AWS_PRIVATE_S3_ACCESS_KEY_ID', setting('AWS_PRIVATE_ACCESS_KEY_ID', self.access_key))
        self.secret_key = setting('AWS_PRIVATE_S3_SECRET_ACCESS_KEY', setting('AWS_PRIVATE_SECRET_ACCESS_KEY', self.secret_key))
        if hasattr(self, "get_default_settings"):
            default_settings = self.get_default_settings()
            self.gzip = setting('AWS_PRIVATE_IS_GZIPPED', default_settings["gzip"])  # fallback to default
            self.url_protocol = setting('AWS_PRIVATE_S3_URL_PROTOCOL', default_settings["url_protocol"])  # fallback to default
            self.region_name = setting('AWS_PRIVATE_S3_REGION_NAME', default_settings["region_name"])  # fallback to default
        else:  # backward compatibility
            self.gzip = setting('AWS_PRIVATE_IS_GZIPPED', self.gzip)
            self.url_protocol = setting('AWS_PRIVATE_S3_URL_PROTOCOL', self.url_protocol)
            self.region_name = setting('AWS_PRIVATE_S3_REGION_NAME', self.region_name)

    def url(self, name, *args, **kwargs):
        if appconfig.PRIVATE_STORAGE_S3_REVERSE_PROXY or not self.querystring_auth:
            # There is no direct URL possible, return our streaming view instead.
            return reverse('serve_private_file', kwargs={'path': name})
        else:
            # The S3Boto3Storage can generate a presigned URL that is temporary available.
            return super().url(name, *args, **kwargs)


@deconstructible
class PrivateEncryptedS3BotoStorage(PrivateS3BotoStorage):
    """
    Enforced encryption for private storage on S3.
    This is a convience option, it can also be implemented
    through :class:`PrivateS3BotoStorage` by using the proper settings.
    """
    encryption = True

    def __init__(self, **settings):
        super().__init__(**settings)
        self.signature_version = self.signature_version or 's3v4'



import os

from django.db import models
from django.utils.text import slugify

from private_storage.fields import PrivateFileField


class SimpleDossier(models.Model):
    file = PrivateFileField()


class UploadToDossier(models.Model):
    file = PrivateFileField(upload_to='UploadToDossier')


class UploadToCallableDossier(models.Model):
    def upload_to(self, filename):
        return 'UploadToCallableDossier/test/' + filename

    file = PrivateFileField(upload_to=upload_to)


class CustomerDossier(models.Model):

    def upload_subfolder(self):
        # self.pk is still None here!
        return [slugify(self.customer)]

    customer = models.CharField(max_length=100)
    file = PrivateFileField(upload_to='CustomerDossier', upload_subfolder=upload_subfolder)


class CustomerDossierJoin(models.Model):

    def upload_subfolder2(self):
        # Slightly incorrect usage by developers, joining locally.
        return os.path.join(slugify(self.customer), 'sub2')

    customer = models.CharField(max_length=100)
    file = PrivateFileField(upload_to='CustomerDossierJoin', upload_subfolder=upload_subfolder2)



import minio

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from minio_storage.policy import Policy
from minio_storage.storage import MinioStorage

from private_storage import appconfig

_NoValue = object()


def default_setting(name, default=_NoValue):
    result = getattr(settings, name, default)
    if result is _NoValue:
        raise ImproperlyConfigured
    else:
        return result


def private_or_default_setting(private_name, default_name, default_value=_NoValue):
    result = getattr(settings, private_name, None)
    if not result:
        return default_setting(default_name, default_value)
    else:
        return result


class PrivateMinioStorage(MinioStorage):
    def __init__(self):
        endpoint = private_or_default_setting("MINIO_PRIVATE_STORAGE_ENDPOINT", "MINIO_STORAGE_ENDPOINT")
        access_key = private_or_default_setting("MINIO_PRIVATE_STORAGE_ACCESS_KEY", "MINIO_STORAGE_ACCESS_KEY")
        secret_key = private_or_default_setting("MINIO_PRIVATE_STORAGE_SECRET_KEY", "MINIO_STORAGE_SECRET_KEY")
        secure = private_or_default_setting("MINIO_PRIVATE_STORAGE_USE_HTTPS", "MINIO_STORAGE_USE_HTTPS", True)

        bucket_name = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_MEDIA_BUCKET_NAME", "MINIO_STORAGE_MEDIA_BUCKET_NAME"
        )
        base_url = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_MEDIA_URL", "MINIO_STORAGE_MEDIA_URL", None
        )
        auto_create_bucket = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_AUTO_CREATE_MEDIA_BUCKET", "MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET", False
        )
        auto_create_policy = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_AUTO_CREATE_MEDIA_POLICY", "MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY", "GET_ONLY"
        )

        policy_type = Policy.get
        if isinstance(auto_create_policy, str):
            policy_type = Policy(auto_create_policy)
            auto_create_policy = True

        presign_urls = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_MEDIA_USE_PRESIGNED", "MINIO_STORAGE_MEDIA_USE_PRESIGNED", False
        )
        backup_format = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_MEDIA_BACKUP_FORMAT", "MINIO_STORAGE_MEDIA_BACKUP_FORMAT", False
        )
        backup_bucket = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_MEDIA_BACKUP_BUCKET", "MINIO_STORAGE_MEDIA_BACKUP_BUCKET", False
        )
        assume_bucket_exists = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_ASSUME_MEDIA_BUCKET_EXISTS", "MINIO_STORAGE_ASSUME_MEDIA_BUCKET_EXISTS", False
        )
        object_metadata = private_or_default_setting(
            "MINIO_PRIVATE_STORAGE_OBJECT_METADATA", "MINIO_STORAGE_MEDIA_OBJECT_METADATA", None
        )

        client = minio.Minio(endpoint, access_key=access_key, secret_key=secret_key, secure=secure)

        super().__init__(
            client,
            bucket_name,
            auto_create_bucket=auto_create_bucket,
            auto_create_policy=auto_create_policy,
            policy_type=policy_type,
            base_url=base_url,
            presign_urls=presign_urls,
            backup_format=backup_format,
            backup_bucket=backup_bucket,
            assume_bucket_exists=assume_bucket_exists,
            object_metadata=object_metadata,
        )

    def get_modified_time(self, name):
        return self.modified_time(name)

    def url(self, name: str, *args, **kwargs) -> str:
        if appconfig.PRIVATE_STORAGE_MINO_REVERSE_PROXY:
            return reverse('serve_private_file', kwargs={'path': name})
        return super().url(name, *args, **kwargs)
